/*--------------------------------------------------------------------------*/
// Sample Application													
//																			
// Chapter 12																
//																			
// Learn DirectX in 24 Hours												
// by Robert Dunlop															
//																			
// Copyright (C) 1999														
/*--------------------------------------------------------------------------*/

//------ Include Files ------//

#include "stdafx.h"
#define D3D_OVERLOADS
#define INITGUID
#include "windef.h"
#include <mmsystem.h>
#include <ddraw.h>
#include <d3d.h>
#include <d3dtypes.h>
#include "d3dutil.h"
#include "d3dmath.h"
#include "cube.h"

#define SafeRelease(x) if (x) { x->Release(); x=NULL;}
#define SafeDelete(x) if (x) {delete x;x=NULL;}

//------ Window Class Information ------//

static char szClass[] = "XmplHr12Class";
static char szCaption[] = "Example - Hour 12";

//------ Global Interface Pointers ------//

LPDIRECTDRAW7 lpDD=NULL;
LPDIRECTDRAWSURFACE7 lpDDSPrimary=NULL;
LPDIRECTDRAWSURFACE7 lpDDSBack=NULL;
LPDIRECT3D7 lpD3D=NULL;
LPDIRECT3DDEVICE7 lpDevice=NULL;

//----- Rotation position and speed -----//

float rotAngle=g_PI;		// current angle
float rotVel=0.0f;			// current velociy of spin

//----- Elevation of viewer and vertical speed -----//

float elevation=350.0f;		// current elevation
float liftVel=0.0f;			// rate of rise / decent

//----- Define distance of viewer orbit from target ------//

#define ORBIT			800.0f

//------ Storage for Cube Objects -----//

#define NUM_ROWS		4
#define NUM_COLUMNS		4

CCube *cubes[NUM_ROWS][NUM_COLUMNS];

//------ Error Return String ------//

const char *ErrStr=NULL;	

//------ Error Messages ------//

const char Err_Reg_Class[]			= "Error Registering Window Class";
const char Err_Create_Win[]			= "Error Creating Window";
const char Err_DirectDrawCreate[]	= "DirectDrawCreate FAILED";
const char Err_Query[]				= "QueryInterface FAILED";
const char Err_Coop[]				= "SetCooperativeLevel FAILED";
const char Err_CreateSurf[]			= "CreateSurface FAILED";
const char Err_DispMode[]			= "Error Setting Display Mode";
const char Err_Device[]				= "Device Creation Failed";
const char Err_SetView[]			= "Viewport settings failed";

//------ Function Prototypes -----//

void Cleanup();
void create_objects();
static BOOL Init(HINSTANCE hInstance, int nCmdShow);
BOOL init_d3d();
BOOL init_ddraw(HWND hWnd);
void render_frame(float elapsed);

// ----- Function Definitions -----//

void Cleanup()
{
	// de-allocate block objects

	for (int i=0;i<NUM_ROWS;i++)
		for (int j=0;j<NUM_COLUMNS;j++)
			SafeDelete(cubes[i][j]);

	// release 3D interfaces

	SafeRelease(lpDevice);
	SafeRelease(lpD3D);
	
	// release DirectDraw interfaces

	SafeRelease(lpDDSPrimary);
	SafeRelease(lpDD);

	// display error if one thrown

	if (ErrStr) {
		MessageBox(NULL, ErrStr, szCaption, MB_OK);
		ErrStr=NULL;
	}

}

LRESULT CALLBACK 
WindowProc(HWND hWnd, unsigned uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg)
    {
		case WM_DESTROY:

			Cleanup();
            PostQuitMessage(0);
            break;

	    case WM_KEYDOWN: 

	        switch (wParam) { 

			case VK_UP:
			
				// move up

				liftVel=200.0f;
				break;

			case VK_DOWN:

				// move down

				liftVel=-200.0f;
				break;

			case VK_RIGHT:

				// rotate to the right

				rotVel=0.5f;
				break;

			case VK_LEFT:

				// rotate to the left

				rotVel=-0.5f;
				break; 
			}
			break;

	    case WM_KEYUP: 

	        switch (wParam) { 

			case VK_UP:
			case VK_DOWN:
			
				liftVel=0.0f;
				break;

			case VK_RIGHT:
			case VK_LEFT:

				// rotate to the left

				rotVel=0.0f;
				break; 

			case VK_ESCAPE:

				// exit the program on escape

				DestroyWindow(hWnd);
				break;

			}
			break;

        default:
            return DefWindowProc(hWnd, uMsg, wParam, lParam);
    }
	
	return 0L;
}

BOOL init_ddraw(HWND hWnd)
{
	HRESULT ddrval;

    // Create the main DirectDraw object

    ddrval = DirectDrawCreateEx(NULL, (void **) &lpDD, IID_IDirectDraw7, NULL);
    if (ddrval != DD_OK) {
		ErrStr=Err_DirectDrawCreate;
		return FALSE;
	}

    // Set our cooperative level

    ddrval = lpDD->SetCooperativeLevel(hWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN );
    if (ddrval != DD_OK) {
		ErrStr=Err_Coop;
		return FALSE;
	}

	// Set the display mode

	ddrval = lpDD->SetDisplayMode( 640, 480, 16, 0, 0);
	if (ddrval !=DD_OK) {
		ErrStr=Err_DispMode;
		return FALSE;
	}

    // Create the primary surface with 1 back buffer

    DDSURFACEDESC2 ddsd;
	ZeroMemory(&ddsd,sizeof(ddsd));
    ddsd.dwSize = sizeof( ddsd );
    ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |
                          DDSCAPS_FLIP | DDSCAPS_3DDEVICE |
                          DDSCAPS_COMPLEX;
    ddsd.dwBackBufferCount = 1;
    ddrval = lpDD->CreateSurface( &ddsd, &lpDDSPrimary, NULL );
	if (ddrval!=DD_OK) {
		ErrStr=Err_CreateSurf;
		return FALSE;
	}

	// Fetch back buffer interface

	DDSCAPS2 ddscaps;
	ZeroMemory(&ddscaps,sizeof(ddscaps));
	ddscaps.dwCaps=DDSCAPS_BACKBUFFER;
	ddrval=lpDDSPrimary->GetAttachedSurface(&ddscaps,&lpDDSBack);
	if (ddrval!=DD_OK) {
		ErrStr=Err_CreateSurf;
		return FALSE;
	}

	// return success to caller

	return TRUE;
}

BOOL init_d3d()
{
	// get master 3D interface

	if (FAILED(lpDD->QueryInterface(IID_IDirect3D7, (LPVOID *)&lpD3D))) {
		ErrStr=Err_Query;
		return FALSE;
	}
	
	// set up the best device interface

	if (lpD3D->CreateDevice(IID_IDirect3DTnLHalDevice,lpDDSBack,&lpDevice)!=D3D_OK) {
		if (lpD3D->CreateDevice(IID_IDirect3DHALDevice,lpDDSBack,&lpDevice)!=D3D_OK) {
			if (lpD3D->CreateDevice(IID_IDirect3DMMXDevice,lpDDSBack,&lpDevice)!=D3D_OK) {
				if (lpD3D->CreateDevice(IID_IDirect3DRGBDevice,lpDDSBack,&lpDevice)!=D3D_OK) {
					ErrStr=Err_Device;
					return FALSE;
				}
			}
		}
	}

	// set up the viewport

	D3DVIEWPORT7 view;
	view.dwX=0;
	view.dwY=0; 
	view.dwWidth=640;
	view.dwHeight=480; 
	view.dvMinZ=0.0f;
	view.dvMaxZ=1.0f;
	if (lpDevice->SetViewport(&view)!=D3D_OK) {
		ErrStr=Err_SetView;
		return FALSE;
	}
	
	D3DMATRIX proj_m;
	D3DUtil_SetProjectionMatrix(proj_m, 0.785f, 1.333f, 10.0f, 2000.0f );
	lpDevice->SetTransform(D3DTRANSFORMSTATE_PROJECTION,&proj_m);

	// disable Direct3D lighting, since we will provide our own

	lpDevice->SetRenderState(D3DRENDERSTATE_LIGHTING,FALSE);

	// return success to caller
	
	return TRUE;
}

void create_objects()
{
	// create cubes on a grid

	for (int i=0;i<NUM_ROWS;i++) 
		for (int j=0;j<NUM_COLUMNS;j++) 
			cubes[i][j]=new CCube(D3DVECTOR(125.0f-100.0f*j,0.0f,125.0f-100.0f*i),
								  D3DVECTOR(50.0f,150.0f,50.0f),
								  0.95f,1.0f,0.8f);
}

void render_frame(float elapsed)
{
	// recover any lost surfaces

	if (lpDDSPrimary->IsLost()==DDERR_SURFACELOST)
		lpDDSPrimary->Restore();
	if (lpDDSBack->IsLost()==DDERR_SURFACELOST)
		lpDDSBack->Restore();

	// increment viewer position

	elevation+=liftVel*elapsed;
	rotAngle+=rotVel*elapsed;

	// calculate current viewer position

	D3DVECTOR view_loc;
	view_loc.y=elevation;				// set elevation
	view_loc.x=sinf(rotAngle)*ORBIT;		// calculate position on x,z plane
	view_loc.z=cosf(rotAngle)*ORBIT;

	// create and set the view matrix

	D3DMATRIX view_matrix;
	D3DUtil_SetViewMatrix(view_matrix,
						  view_loc,
						  D3DVECTOR(0.0f,0.0f,0.0f),
						  D3DVECTOR(0.0f,1.0f,0.0f));
	lpDevice->SetTransform(D3DTRANSFORMSTATE_VIEW,&view_matrix);

 	// clear the viewport

    lpDevice->Clear(0,NULL,D3DCLEAR_TARGET,0,1.0f,0);

	// start the scene render

	if( SUCCEEDED( lpDevice->BeginScene() ) ) {

		// loop through the cubes and draw

		for (int i=0;i<NUM_ROWS;i++) 
			for (int j=0;j<NUM_COLUMNS;j++) 
				if (cubes[i][j])
					cubes[i][j]->draw(lpDevice);

		// end the scene

		lpDevice->EndScene();
	}

	// flip to the primary surface

	lpDDSPrimary->Flip(0,DDFLIP_WAIT);
}

//------ Function to Initialize DirectDraw and the Application ------//

static BOOL Init(HINSTANCE hInstance, int nCmdShow)
{
    WNDCLASS                    wc;

    // Set up and register window class

    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = (WNDPROC) WindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = sizeof(DWORD);
    wc.hInstance = hInstance;
    wc.hIcon = NULL;
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH) GetStockObject(BLACK_BRUSH);
    wc.lpszMenuName = NULL;
    wc.lpszClassName = szClass;
    if (!RegisterClass(&wc)) {
		ErrStr=Err_Reg_Class;
        return FALSE;
	}

    // Get dimensions of display

    int ScreenWidth = GetSystemMetrics(SM_CXSCREEN);
    int ScreenHeight = GetSystemMetrics(SM_CYSCREEN);

    // Create a window and display

	HWND hWnd;

    hWnd = CreateWindow(szClass,							// class
                        szCaption,							// caption
						WS_VISIBLE|WS_POPUP,				// style 
						0,									// left
						0,									// top
						ScreenWidth,						// width
						ScreenHeight,						// height
                        NULL,								// parent window
                        NULL,								// menu 
                        hInstance,							// instance
                        NULL);								// parms
    if (!hWnd) {
		ErrStr=Err_Create_Win;
        return FALSE;
	}
    ShowWindow(hWnd, nCmdShow);
    UpdateWindow(hWnd);

	// initialze DirectDraw

	if (!init_ddraw(hWnd)) return FALSE;

	// initialize Direct3D

	if (!init_d3d()) return FALSE;

	// create 3D objects

	create_objects();

	// return success to caller

	return TRUE;
}

//------ Application Loop ------//

int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
	LONGLONG cur_time;       // current time
	LONGLONG perf_cnt;       // performance timer frequency
	BOOL perf_flag=FALSE;    // flag determining which timer to use
	LONGLONG last_time=0;	 // time of previous frame
	float time_elapsed;	 // time since previous frame
	float time_scale;		 // scaling factor for time

	// initialize the application, exit on failure

    if (!Init(hInstance, nCmdShow)) {
		Cleanup();
        return FALSE;
	}

	// is there a performance counter available? 

	if (QueryPerformanceFrequency((LARGE_INTEGER *) &perf_cnt)) { 

		// yes, set timer info and get starting time

		perf_flag=TRUE;
		QueryPerformanceCounter((LARGE_INTEGER *) &last_time); 
		time_scale=1.0f/perf_cnt;

	} else { 

		// no performance counter, read in using timeGetTime 

		last_time=timeGetTime(); 
		time_scale=0.001f;
	} 

    // Now we're ready to recieve and process Windows messages.

    BOOL bGotMsg;
    MSG  msg;
    PeekMessage( &msg, NULL, 0U, 0U, PM_NOREMOVE );
	
    while( WM_QUIT != msg.message  )
    {
        bGotMsg = PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE );
        if( bGotMsg )
        {
			TranslateMessage( &msg );
            DispatchMessage( &msg );
        } else {
			
			// use the appropriate method to get time 
			// and calculate elapsed time since last frame

			if (perf_flag) 
				QueryPerformanceCounter((LARGE_INTEGER *) &cur_time); 
			else 
				cur_time=timeGetTime(); 

			// calculate elapsed time

			time_elapsed=(cur_time-last_time)*time_scale;

			// save frame time

			last_time=cur_time;
				
			// render the frame
			
			render_frame(time_elapsed);
		}
    }

	// return final message

    return msg.wParam;
}


